/*
 * Copyright (C) 2016  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <libc64_dsp/include/inc_overlay.h>
#include <stdint.h>
#include "tiles_dsp.h"
#include "idma.h"

typedef struct {
    void *src, *dst;
    uint16_t width, height;
    uint16_t maxheight, dstx;
    uint16_t srcpitch, dstpitch;
    uint8_t prepared, alpha;
    uint8_t transfering_sources;
    uint8_t reserved[9];
} transfer_data;

void tiles_init(dsp_screen *data)
{
    syscalls.qdma_init();
    syscalls.cache_inv(data->surfacedata, data->surfacedata_size, 1);
    syscalls.cache_inv(data->pixels[0], data->pitch * data->h, 1);
    syscalls.cache_inv(data->pixels[1], data->pitch * data->h, 1);
}

static void blit_alpha_32_16_line(const uint32_t * restrict srcalpha, const uint16_t * restrict srcscreen, uint16_t * restrict dstscreen, uint32_t width)
{
    _nassert((int)srcalpha % 8 == 0);
    _nassert((int)srcscreen % 8 == 0);
    _nassert((int)dstscreen % 8 == 0);

    for (; width != 0; width--)
    {
        uint32_t srcpix = _amem4_const(srcalpha);
        srcalpha++;
        uint16_t screenpix = _amem2_const(srcscreen);
        srcscreen++;

        int alpha = (srcpix >> 24);

        // convert screen pixel from 565 to 888 format
        int r = (screenpix >> 8) & 0xf8;
        r = r | (r >> 5);
        int g = (screenpix >> 3) & 0xfc;
        g = g | (g >> 6);
        int b = (screenpix << 3) & 0xf8;
        b = b | (b >> 5);

        r = ((((((srcpix >> 16) & 0xff) - r) * alpha) + 255) >> 8) + r;
        g = ((((((srcpix >>  8) & 0xff) - g) * alpha) + 255) >> 8) + g;
        b = ((((( srcpix        & 0xff) - b) * alpha) + 255) >> 8) + b;

        _amem2(dstscreen) = ((r & 0xf8) << 8) | ((g & 0xfc) << 3) | (b >> 3);
        dstscreen++;
    }
}

static void blit_alpha_32_16_line_color(uint32_t srcpix, const uint16_t * restrict srcscreen, uint16_t * restrict dstscreen, uint32_t width)
{
    _nassert((int)srcscreen % 8 == 0);
    _nassert((int)dstscreen % 8 == 0);

    for (; width != 0; width--)
    {
        uint16_t screenpix = _amem2_const(srcscreen);
        srcscreen++;

        int alpha = (srcpix >> 24);

        // convert screen pixel from 565 to 888 format
        int r = (screenpix >> 8) & 0xf8;
        r = r | (r >> 5);
        int g = (screenpix >> 3) & 0xfc;
        g = g | (g >> 6);
        int b = (screenpix << 3) & 0xf8;
        b = b | (b >> 5);

        r = ((((((srcpix >> 16) & 0xff) - r) * alpha) + 255) >> 8) + r;
        g = ((((((srcpix >>  8) & 0xff) - g) * alpha) + 255) >> 8) + g;
        b = ((((( srcpix        & 0xff) - b) * alpha) + 255) >> 8) + b;

        _amem2(dstscreen) = ((r & 0xf8) << 8) | ((g & 0xfc) << 3) | (b >> 3);
        dstscreen++;
    }
}

// linked or parallel transfers ?
// parallel transfers seem to be a bit faster in this case
#define LINKED_TRANSFERS 0

#define MIN_HEIGHT(transfer) ( (transfer->height < transfer->maxheight)?transfer->height:transfer->maxheight )

static int transfer_dependency(transfer_data * restrict next_transfer, transfer_data * restrict current_transfer)
{
    if (next_transfer->dstx < current_transfer->dstx)
    {
        if (next_transfer->dstx + next_transfer->width < current_transfer->dstx) return 0;
    }
    else
    {
        if (current_transfer->dstx + current_transfer->width < next_transfer->dstx) return 0;
    }

    if ((uintptr_t)next_transfer->dst < (uintptr_t)current_transfer->dst)
    {
        if ((next_transfer->dstpitch * MIN_HEIGHT(next_transfer)) + (uintptr_t)next_transfer->dst < (uintptr_t)current_transfer->dst) return 0;
        return 1;
    }
    else
    {
        if ((current_transfer->dstpitch * MIN_HEIGHT(current_transfer)) + (uintptr_t)current_transfer->dst < (uintptr_t)next_transfer->dst) return 0;
        return 1;
    }
}

/*
 * 1 pixel:
 * 2 * (srcalpha + srcscreen + dstscreen)
 * 2 * (4 + 2 + 2)
 * 16 bytes per pixes
 * (24*1024-64-32-32) / 16 = max 1528 pixels at once
 */
#define MAX_PIXELS_PER_LINE 1528
// should be divisible by 4

static void draw_line_16(const dsp_screen * restrict data, const cmd_params * restrict cmd)
{
    uint32_t color4 = (cmd->w << 16) | cmd->h;
    uint16_t color = ((color4 & 0xf80000) >> 8) | ((color4 & 0xfc00) >> 5) | ((color4 & 0xf8) >> 3);

    uint32_t dstpitch = data->pitch;
    uint16_t *dst = (uint16_t *) ( (dstpitch * cmd->srcy) + (2 * cmd->srcx) + (uintptr_t) data->pixels[cmd->dst_screen] );

    int adiffx = cmd->dstx - cmd->srcx;
    if (adiffx < 0) adiffx = -adiffx;
    if (adiffx <= (cmd->dsty - cmd->srcy))
    {
        int h = cmd->dsty - cmd->srcy;
        int32_t xdiff = ((cmd->dstx - cmd->srcx) << 16) / h;
        int32_t xvalue = 0;

        for (; h != 0; h--)
        {
            *dst = color;
            syscalls.cache_wbInv(dst, 1, 0);

            dst = (uint16_t *) (dstpitch + (uintptr_t)dst);
            xvalue += xdiff;
            if (xvalue & 0xffff0000)
            {
                dst += xvalue >> 16;
                xvalue &= 0xffff;
            }
        }
        syscalls.cache_wait();
    }
    else if (cmd->srcx <= cmd->dstx)
    {
        int w = cmd->dstx - cmd->srcx;
        int32_t ydiff = ((cmd->dsty - cmd->srcy) << 16) / w;
        int32_t yvalue = 0;
        void *first = dst;

        for (; w != 0; w--)
        {
            *dst = color;

            yvalue += ydiff;
            if (yvalue & 0xffff0000)
            {
                syscalls.cache_wbInv(first, 1 + ((uintptr_t)dst) - (uintptr_t)first, 0);

                dst = (uint16_t *) (dstpitch + (uintptr_t)dst);
                first = dst + 1;
                yvalue &= 0xffff;
            }
            dst ++;
        }
        if (first != dst)
        {
            syscalls.cache_wbInv(first, ((uintptr_t)dst) - (uintptr_t)first, 0);
        }
        syscalls.cache_wait();
    }
    else
    {
        int w = cmd->srcx - cmd->dstx;
        int32_t ydiff = ((cmd->dsty - cmd->srcy) << 16) / w;
        int32_t yvalue = 0;
        void *last = dst;

        for (; w != 0; w--)
        {
            *dst = color;

            yvalue += ydiff;
            if (yvalue & 0xffff0000)
            {
                syscalls.cache_wbInv(dst, 1 + ((uintptr_t)last) - (uintptr_t)dst, 0);

                dst = (uint16_t *) (dstpitch + (uintptr_t)dst);
                last = dst - 1;
                yvalue &= 0xffff;
            }
            dst --;
        }
        if (last != dst)
        {
            syscalls.cache_wbInv(dst, ((uintptr_t)last) - (uintptr_t)dst, 0);
        }
        syscalls.cache_wait();
    }
}

static void prepare_transfer_16(const dsp_screen * restrict data, const cmd_params * restrict cmd, transfer_data * restrict transfer)
{
    dsp_surface * restrict surface = (dsp_surface *) (data->surfaces_base + cmd->src_offset);
    transfer->src = (void *) ( (surface->pitch * cmd->srcy) + (surface->bytes_per_pixel * cmd->srcx) + (uintptr_t) surface->pixels );
    transfer->dst = (void *) ( (data->pitch * cmd->dsty) + (2 * cmd->dstx) + (uintptr_t) data->pixels[cmd->dst_screen] );
    transfer->width = cmd->w;
    transfer->height = cmd->h;
    transfer->dstx = cmd->dstx;
    transfer->srcpitch = surface->pitch;
    transfer->dstpitch = data->pitch;
    transfer->prepared = 1;
    transfer->alpha = surface->has_alpha;
    transfer->transfering_sources = 0;
    if ((!surface->has_alpha) || (cmd->w * cmd->h <= MAX_PIXELS_PER_LINE))
    {
        transfer->maxheight = cmd->h;
    }
    else
    {
        transfer->maxheight = MAX_PIXELS_PER_LINE / cmd->w;
    }
}

static void tiles_run_16(dsp_screen *data)
{
    cmd_params * restrict cmd_queue;

    transfer_data * restrict current_transfer;
    transfer_data * restrict next_transfer;

    uint32_t * restrict srcalpha1;
    uint32_t * restrict srcalpha2;

    uint16_t * restrict srcscreen1;
    uint16_t * restrict srcscreen2;

    uint16_t * restrict dstscreen1;
    uint16_t * restrict dstscreen2;

    cmd_queue = (cmd_params *) (IVA2_L1DSRAM_BASE_DSP + DSP_L1SRAM_FSHM_BASE_OFF);

    current_transfer = (transfer_data *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP);
    next_transfer = current_transfer + 1;

    current_transfer->prepared = 0;

    srcalpha1 = (uint32_t *) (next_transfer + 1);
    srcalpha2 = srcalpha1 + MAX_PIXELS_PER_LINE;

    srcscreen1 = (uint16_t *) (srcalpha2 + MAX_PIXELS_PER_LINE);
    srcscreen2 = srcscreen1 + MAX_PIXELS_PER_LINE;

    dstscreen1 = srcscreen2 + MAX_PIXELS_PER_LINE;
    dstscreen2 = dstscreen1 + MAX_PIXELS_PER_LINE;

    int transfering_result = 0;

    int read_index = data->cmd_queue_read;
    while (1)
    {
        while (read_index == data->cmd_queue_write)
        {
            if (data->quit) return;
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
        };

        do {
            switch (cmd_queue[read_index].type)
            {
                case PARAM_FLUSH:
                    if (transfering_result)
                    {
                        transfering_result = 0;
                        syscalls.qdma_wait(QDMA_0);
                    }
                    break;
                case PARAM_BLIT:
                    {
                        if (!current_transfer->prepared)
                        {
                            prepare_transfer_16(data, &(cmd_queue[read_index]), current_transfer);
                        }

                        /*
                         * A) copy dst to localsrcscreen
                         * B) copy src to localsrcalpha
                         * C) set localdstscreen using localsrcalpha and localsrcscreen (depends on A,B)
                         * D) copy localdstscreen to dst
                        **/
                        if (current_transfer->alpha)
                        {
                            // first line:
                            // if (CURR A,B)
                            //   // nothing
                            if (!current_transfer->transfering_sources)
                            {
                                // else if (PREV D)
                                if (transfering_result)
                                {
                                    //   first B (Q1)
                                    syscalls.qdma_copy2d(
                                        QDMA_1,                         // sUI _qdmaNr
                                        srcalpha1,                      // void *_dst
                                        current_transfer->width * 4,    // sS32 _dstPitch
                                        current_transfer->src,          // const void *_src
                                        current_transfer->srcpitch,     // sS32 _srcPitch
                                        current_transfer->width * 4,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                    //   finish PREV D (Q0)
                                    syscalls.qdma_wait(QDMA_0);
                                    //   first A (Q2)
                                    syscalls.qdma_copy2d(
                                        QDMA_2,                         // sUI _qdmaNr
                                        srcscreen1,                     // void *_dst
                                        current_transfer->width * 2,    // sS32 _dstPitch
                                        current_transfer->dst,          // const void *_src
                                        current_transfer->dstpitch,     // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                }
                                else
                                {
                                    // else
                                    //   first B,A (linked or parallel) (Q2 or Q1,Q2)
                                    if (LINKED_TRANSFERS)
                                    {
                                        // A
                                        syscalls.qdma_link2d(
                                            QDMA_2,                         // sUI _qdmaNr
                                            8,                              // sUI _paramNr
                                            0xffff,                         // sUI _linkToParamNr
                                            srcscreen1,                     // void *_dst
                                            current_transfer->width * 2,    // sS32 _dstPitch
                                            current_transfer->dst,          // const void *_src
                                            current_transfer->dstpitch,     // sS32 _srcPitch
                                            current_transfer->width * 2,    // sU32 _numBytesX
                                            MIN_HEIGHT(current_transfer)    // sU32 _numY
                                        );
                                        // B
                                        syscalls.qdma_link2d(
                                            QDMA_2,                         // sUI _qdmaNr
                                            QDMA_2,                         // sUI _paramNr
                                            8,                              // sUI _linkToParamNr
                                            srcalpha1,                      // void *_dst
                                            current_transfer->width * 4,    // sS32 _dstPitch
                                            current_transfer->src,          // const void *_src
                                            current_transfer->srcpitch,     // sS32 _srcPitch
                                            current_transfer->width * 4,    // sU32 _numBytesX
                                            MIN_HEIGHT(current_transfer)    // sU32 _numY
                                        );
                                    }
                                    else
                                    {
                                        // B
                                        syscalls.qdma_copy2d(
                                            QDMA_1,                         // sUI _qdmaNr
                                            srcalpha1,                      // void *_dst
                                            current_transfer->width * 4,    // sS32 _dstPitch
                                            current_transfer->src,          // const void *_src
                                            current_transfer->srcpitch,     // sS32 _srcPitch
                                            current_transfer->width * 4,    // sU32 _numBytesX
                                            MIN_HEIGHT(current_transfer)    // sU32 _numY
                                        );
                                        // A
                                        syscalls.qdma_copy2d(
                                            QDMA_2,                         // sUI _qdmaNr
                                            srcscreen1,                     // void *_dst
                                            current_transfer->width * 2,    // sS32 _dstPitch
                                            current_transfer->dst,          // const void *_src
                                            current_transfer->dstpitch,     // sS32 _srcPitch
                                            current_transfer->width * 2,    // sU32 _numBytesX
                                            MIN_HEIGHT(current_transfer)    // sU32 _numY
                                        );
                                    }
                                }
                            }
                            transfering_result = 1;

                            // line loop:
                            while (current_transfer->height > current_transfer->maxheight)
                            {
                                current_transfer->src = (void *) (current_transfer->srcpitch + (uintptr_t) current_transfer->src);
                                current_transfer->height -= current_transfer->maxheight;

                                // finish current B,A (Q1,Q2)
                                syscalls.qdma_wait(QDMA_1);
                                syscalls.qdma_wait(QDMA_2);
                                // next B,A (linked or parallel) (Q2 or Q1,Q2) | current C
                                if (LINKED_TRANSFERS)
                                {
                                    // A
                                    syscalls.qdma_link2d(
                                        QDMA_2,                         // sUI _qdmaNr
                                        9,                              // sUI _paramNr
                                        0xffff,                         // sUI _linkToParamNr
                                        srcscreen2,                     // void *_dst
                                        current_transfer->width * 2,    // sS32 _dstPitch
                                        (void *) (current_transfer->dstpitch + (uintptr_t) current_transfer->dst),
                                                                        // const void *_src
                                        current_transfer->dstpitch,     // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                    // B
                                    syscalls.qdma_link2d(
                                        QDMA_2,                         // sUI _qdmaNr
                                        QDMA_2,                         // sUI _paramNr
                                        9,                              // sUI _linkToParamNr
                                        srcalpha2,                      // void *_dst
                                        current_transfer->width * 4,    // sS32 _dstPitch
                                        current_transfer->src,          // const void *_src
                                        current_transfer->srcpitch,     // sS32 _srcPitch
                                        current_transfer->width * 4,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                }
                                else
                                {
                                    // B
                                    syscalls.qdma_copy2d(
                                        QDMA_1,                         // sUI _qdmaNr
                                        srcalpha2,                      // void *_dst
                                        current_transfer->width * 4,    // sS32 _dstPitch
                                        current_transfer->src,          // const void *_src
                                        current_transfer->srcpitch,     // sS32 _srcPitch
                                        current_transfer->width * 4,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                    // A
                                    syscalls.qdma_copy2d(
                                        QDMA_2,                         // sUI _qdmaNr
                                        srcscreen2,                     // void *_dst
                                        current_transfer->width * 2,    // sS32 _dstPitch
                                        (void *) (current_transfer->dstpitch + (uintptr_t) current_transfer->dst),
                                                                        // const void *_src
                                        current_transfer->dstpitch,     // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        MIN_HEIGHT(current_transfer)    // sU32 _numY
                                    );
                                }
                                blit_alpha_32_16_line(srcalpha1, srcscreen1, dstscreen1, current_transfer->maxheight * current_transfer->width);

                                // finish previous D (Q0)
                                syscalls.qdma_wait(QDMA_0);
                                // current D (Q0)
                                syscalls.qdma_copy2d(
                                    QDMA_0,                         // sUI _qdmaNr
                                    current_transfer->dst,          // void *_dst
                                    current_transfer->dstpitch,     // sS32 _dstPitch
                                    dstscreen1,                     // const void *_src
                                    current_transfer->width * 2,    // sS32 _srcPitch
                                    current_transfer->width * 2,    // sU32 _numBytesX
                                    current_transfer->maxheight     // sU32 _numY
                                );

                                current_transfer->dst = (void *) (current_transfer->dstpitch + (uintptr_t) current_transfer->dst);
                                {
                                    uint32_t * restrict srcalpha;
                                    srcalpha = srcalpha1;
                                    srcalpha1 = srcalpha2;
                                    srcalpha2 = srcalpha;
                                }
                                {
                                    uint16_t * restrict srcscreen;
                                    srcscreen = srcscreen1;
                                    srcscreen1 = srcscreen2;
                                    srcscreen2 = srcscreen;
                                }
                                {
                                    uint16_t * restrict dstscreen;
                                    dstscreen = dstscreen1;
                                    dstscreen1 = dstscreen2;
                                    dstscreen2 = dstscreen;
                                }
                            }

                            // last line:
                            // finish current B,A (Q1,Q2)
                            syscalls.qdma_wait(QDMA_1);
                            syscalls.qdma_wait(QDMA_2);

                            int next_read_index = (read_index + 1) & (CMD_QUEUE_LENGTH - 1);
                            if ((next_read_index != data->cmd_queue_write) && (cmd_queue[next_read_index].type == PARAM_BLIT))
                            {
                                prepare_transfer_16(data, &(cmd_queue[next_read_index]), next_transfer);
                            }
                            else
                            {
                                next_transfer->prepared = 0;
                            }

                            // if (! NEXT A,B)
                            if ((!next_transfer->prepared) || (!next_transfer->alpha))
                            {
                                //   current C
                                blit_alpha_32_16_line(srcalpha1, srcscreen1, dstscreen1, current_transfer->height * current_transfer->width);
                                //   finish previous D (Q0)
                                syscalls.qdma_wait(QDMA_0);
                                //   current D (Q0) (must finish on first line of next surface) (??? link NEXT NONALPHA ???)
                                syscalls.qdma_copy2d(
                                    QDMA_0,                         // sUI _qdmaNr
                                    current_transfer->dst,          // void *_dst
                                    current_transfer->dstpitch,     // sS32 _dstPitch
                                    dstscreen1,                     // const void *_src
                                    current_transfer->width * 2,    // sS32 _srcPitch
                                    current_transfer->width * 2,    // sU32 _numBytesX
                                    current_transfer->height        // sU32 _numY
                                );
                            }
                            else
                            {
                                next_transfer->transfering_sources = 1;
                                // else if (NEXT A depends on CURR D)
                                if (transfer_dependency(next_transfer, current_transfer))
                                {
                                    //   NEXT B (Q1) | current C
                                    syscalls.qdma_copy2d(
                                        QDMA_1,                     // sUI _qdmaNr
                                        srcalpha2,                  // void *_dst
                                        next_transfer->width * 4,   // sS32 _dstPitch
                                        next_transfer->src,         // const void *_src
                                        next_transfer->srcpitch,    // sS32 _srcPitch
                                        next_transfer->width * 4,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                    blit_alpha_32_16_line(srcalpha1, srcscreen1, dstscreen1, current_transfer->height * current_transfer->width);
                                    //   finish previous D (Q0)
                                    syscalls.qdma_wait(QDMA_0);
                                    //   current D, NEXT A (Q2) (linked !!!) (must finish on first line of next surface)
                                    // A
                                    syscalls.qdma_link2d(
                                        QDMA_2,                     // sUI _qdmaNr
                                        10,                         // sUI _paramNr
                                        0xffff,                     // sUI _linkToParamNr
                                        srcscreen2,                 // void *_dst
                                        next_transfer->width * 2,   // sS32 _dstPitch
                                        next_transfer->dst,         // const void *_src
                                        next_transfer->dstpitch,    // sS32 _srcPitch
                                        next_transfer->width * 2,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                    // D
                                    syscalls.qdma_link2d(
                                        QDMA_2,                         // sUI _qdmaNr
                                        QDMA_2,                         // sUI _paramNr
                                        10,                             // sUI _linkToParamNr
                                        current_transfer->dst,          // void *_dst
                                        current_transfer->dstpitch,     // sS32 _dstPitch
                                        dstscreen1,                     // const void *_src
                                        current_transfer->width * 2,    // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        current_transfer->height    // sU32 _numY
                                    );
                                }
                                else
                                {
                                    // else
                                    //   NEXT B,A (linked or parallel) (Q2 or Q1,Q2) | current C
                                    if (LINKED_TRANSFERS)
                                    {
                                        // A
                                        syscalls.qdma_link2d(
                                            QDMA_2,                     // sUI _qdmaNr
                                            11,                         // sUI _paramNr
                                            0xffff,                     // sUI _linkToParamNr
                                            srcscreen2,                 // void *_dst
                                            next_transfer->width * 2,   // sS32 _dstPitch
                                            next_transfer->dst,         // const void *_src
                                            next_transfer->dstpitch,    // sS32 _srcPitch
                                            next_transfer->width * 2,   // sU32 _numBytesX
                                            MIN_HEIGHT(next_transfer)   // sU32 _numY
                                        );
                                        // B
                                        syscalls.qdma_link2d(
                                            QDMA_2,                     // sUI _qdmaNr
                                            QDMA_2,                     // sUI _paramNr
                                            11,                         // sUI _linkToParamNr
                                            srcalpha2,                  // void *_dst
                                            next_transfer->width * 4,   // sS32 _dstPitch
                                            next_transfer->src,         // const void *_src
                                            next_transfer->srcpitch,    // sS32 _srcPitch
                                            next_transfer->width * 4,   // sU32 _numBytesX
                                            MIN_HEIGHT(next_transfer)   // sU32 _numY
                                        );
                                    }
                                    else
                                    {
                                        // B
                                        syscalls.qdma_copy2d(
                                            QDMA_1,                     // sUI _qdmaNr
                                            srcalpha2,                  // void *_dst
                                            next_transfer->width * 4,   // sS32 _dstPitch
                                            next_transfer->src,         // const void *_src
                                            next_transfer->srcpitch,    // sS32 _srcPitch
                                            next_transfer->width * 4,   // sU32 _numBytesX
                                            MIN_HEIGHT(next_transfer)   // sU32 _numY
                                        );
                                        // A
                                        syscalls.qdma_copy2d(
                                            QDMA_2,                     // sUI _qdmaNr
                                            srcscreen2,                 // void *_dst
                                            next_transfer->width * 2,   // sS32 _dstPitch
                                            next_transfer->dst,         // const void *_src
                                            next_transfer->dstpitch,    // sS32 _srcPitch
                                            next_transfer->width * 2,   // sU32 _numBytesX
                                            MIN_HEIGHT(next_transfer)   // sU32 _numY
                                        );
                                    }
                                    blit_alpha_32_16_line(srcalpha1, srcscreen1, dstscreen1, current_transfer->height * current_transfer->width);
                                    //   finish previous D (Q0)
                                    syscalls.qdma_wait(QDMA_0);
                                    //   current D (Q0) (must finish on first line of next surface)
                                    syscalls.qdma_copy2d(
                                        QDMA_0,                         // sUI _qdmaNr
                                        current_transfer->dst,          // void *_dst
                                        current_transfer->dstpitch,     // sS32 _dstPitch
                                        dstscreen1,                     // const void *_src
                                        current_transfer->width * 2,    // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        current_transfer->height        // sU32 _numY
                                    );
                                }
                            }

                            {
                                uint32_t * restrict srcalpha;
                                srcalpha = srcalpha1;
                                srcalpha1 = srcalpha2;
                                srcalpha2 = srcalpha;
                            }
                            {
                                uint16_t * restrict srcscreen;
                                srcscreen = srcscreen1;
                                srcscreen1 = srcscreen2;
                                srcscreen2 = srcscreen;
                            }
                            {
                                uint16_t * restrict dstscreen;
                                dstscreen = dstscreen1;
                                dstscreen1 = dstscreen2;
                                dstscreen2 = dstscreen;
                            }
                        }
                        else
                        {
                            // if (PREV D)
                            //   finish PREV D (Q0)
                            if (transfering_result)
                            {
                                syscalls.qdma_wait(QDMA_0);
                            }
                            transfering_result = 1;

                            int next_read_index = (read_index + 1) & (CMD_QUEUE_LENGTH - 1);
                            if ((next_read_index != data->cmd_queue_write) && (cmd_queue[next_read_index].type == PARAM_BLIT))
                            {
                                prepare_transfer_16(data, &(cmd_queue[next_read_index]), next_transfer);
                            }
                            else
                            {
                                next_transfer->prepared = 0;
                            }

                            // if (! NEXT A,B)
                            if ((!next_transfer->prepared) || (!next_transfer->alpha))
                            {
                                //   current D (??? link NEXT NONALPHA ???) (Q0)
                                syscalls.qdma_copy2d(
                                    QDMA_0,                         // sUI _qdmaNr
                                    current_transfer->dst,          // void *_dst
                                    current_transfer->dstpitch,     // sS32 _dstPitch
                                    current_transfer->src,          // const void *_src
                                    current_transfer->srcpitch,     // sS32 _srcPitch
                                    current_transfer->width * 2,    // sU32 _numBytesX
                                    current_transfer->height        // sU32 _numY
                                );
                            }
                            else
                            {
                                next_transfer->transfering_sources = 1;
                                // else if (NEXT A depends on CURR D)
                                if (transfer_dependency(next_transfer, current_transfer))
                                {
                                    //   NEXT B (Q1) | current D, NEXT A (Q2) (linked !!!)
                                    // B
                                    syscalls.qdma_copy2d(
                                        QDMA_1,                     // sUI _qdmaNr
                                        srcalpha1,                  // void *_dst
                                        next_transfer->width * 4,   // sS32 _dstPitch
                                        next_transfer->src,         // const void *_src
                                        next_transfer->srcpitch,    // sS32 _srcPitch
                                        next_transfer->width * 4,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                    // A
                                    syscalls.qdma_link2d(
                                        QDMA_2,                     // sUI _qdmaNr
                                        12,                         // sUI _paramNr
                                        0xffff,                     // sUI _linkToParamNr
                                        srcscreen1,                 // void *_dst
                                        next_transfer->width * 2,   // sS32 _dstPitch
                                        next_transfer->dst,         // const void *_src
                                        next_transfer->dstpitch,    // sS32 _srcPitch
                                        next_transfer->width * 2,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                    // D
                                    syscalls.qdma_link2d(
                                        QDMA_2,                     // sUI _qdmaNr
                                        QDMA_2,                     // sUI _paramNr
                                        12,                         // sUI _linkToParamNr
                                        current_transfer->dst,          // void *_dst
                                        current_transfer->dstpitch,     // sS32 _dstPitch
                                        current_transfer->src,          // const void *_src
                                        current_transfer->srcpitch,     // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        current_transfer->height        // sU32 _numY
                                    );
                                }
                                else
                                {
                                    // else
                                    //   current D (Q0) | NEXT B,A (Q2) (linked !!!)
                                    // D
                                    syscalls.qdma_copy2d(
                                        QDMA_0,                         // sUI _qdmaNr
                                        current_transfer->dst,          // void *_dst
                                        current_transfer->dstpitch,     // sS32 _dstPitch
                                        current_transfer->src,          // const void *_src
                                        current_transfer->srcpitch,     // sS32 _srcPitch
                                        current_transfer->width * 2,    // sU32 _numBytesX
                                        current_transfer->height        // sU32 _numY
                                    );
                                    // A
                                    syscalls.qdma_link2d(
                                        QDMA_2,                     // sUI _qdmaNr
                                        13,                         // sUI _paramNr
                                        0xffff,                     // sUI _linkToParamNr
                                        srcscreen1,                 // void *_dst
                                        next_transfer->width * 2,   // sS32 _dstPitch
                                        next_transfer->dst,         // const void *_src
                                        next_transfer->dstpitch,    // sS32 _srcPitch
                                        next_transfer->width * 2,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                    // B
                                    syscalls.qdma_link2d(
                                        QDMA_2,                     // sUI _qdmaNr
                                        QDMA_2,                     // sUI _paramNr
                                        13,                         // sUI _linkToParamNr
                                        srcalpha1,                  // void *_dst
                                        next_transfer->width * 4,   // sS32 _dstPitch
                                        next_transfer->src,         // const void *_src
                                        next_transfer->srcpitch,    // sS32 _srcPitch
                                        next_transfer->width * 4,   // sU32 _numBytesX
                                        MIN_HEIGHT(next_transfer)   // sU32 _numY
                                    );
                                }
                            }
                        }


                        // exchange next_transfer and current_transfer
                        {
                            transfer_data * restrict transfer;
                            transfer = current_transfer;
                            current_transfer = next_transfer;
                            next_transfer = transfer;
                        }
                    }
                    break;
                case PARAM_FRECT:
                    {
                        uint32_t color = (cmd_queue[read_index].src_offset << 16) | cmd_queue[read_index].srcx;

                        if ((color >> 24) != 0xff)
                        {
                            // alpha
                            current_transfer->dst = (void *) ( (data->pitch * cmd_queue[read_index].dsty) + (2 * cmd_queue[read_index].dstx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] );
                            current_transfer->width = cmd_queue[read_index].w;
                            current_transfer->height = cmd_queue[read_index].h;
                            current_transfer->dstpitch = data->pitch;
                            if (cmd_queue[read_index].w * cmd_queue[read_index].h <= MAX_PIXELS_PER_LINE)
                            {
                                current_transfer->maxheight = cmd_queue[read_index].h;
                            }
                            else
                            {
                                current_transfer->maxheight = MAX_PIXELS_PER_LINE / cmd_queue[read_index].w;
                            }

                            if (transfering_result)
                            {
                                syscalls.qdma_wait(QDMA_0);
                            }
                            transfering_result = 1;

                            //   first A (Q2)
                            syscalls.qdma_copy2d(
                                QDMA_2,                         // sUI _qdmaNr
                                srcscreen1,                     // void *_dst
                                current_transfer->width * 2,    // sS32 _dstPitch
                                current_transfer->dst,          // const void *_src
                                current_transfer->dstpitch,     // sS32 _srcPitch
                                current_transfer->width * 2,    // sU32 _numBytesX
                                MIN_HEIGHT(current_transfer)    // sU32 _numY
                            );

                            // line loop:
                            while (current_transfer->height > current_transfer->maxheight)
                            {
                                current_transfer->height -= current_transfer->maxheight;

                                // finish current A (Q2)
                                syscalls.qdma_wait(QDMA_2);
                                // next A (Q2) | current C
                                // A
                                syscalls.qdma_copy2d(
                                    QDMA_2,                         // sUI _qdmaNr
                                    srcscreen2,                     // void *_dst
                                    current_transfer->width * 2,    // sS32 _dstPitch
                                    (void *) (current_transfer->dstpitch + (uintptr_t) current_transfer->dst),
                                                                    // const void *_src
                                    current_transfer->dstpitch,     // sS32 _srcPitch
                                    current_transfer->width * 2,    // sU32 _numBytesX
                                    MIN_HEIGHT(current_transfer)    // sU32 _numY
                                );
                                blit_alpha_32_16_line_color(color, srcscreen1, dstscreen1, current_transfer->maxheight * current_transfer->width);

                                // finish previous D (Q0)
                                syscalls.qdma_wait(QDMA_0);
                                // current D (Q0)
                                syscalls.qdma_copy2d(
                                    QDMA_0,                         // sUI _qdmaNr
                                    current_transfer->dst,          // void *_dst
                                    current_transfer->dstpitch,     // sS32 _dstPitch
                                    dstscreen1,                     // const void *_src
                                    current_transfer->width * 2,    // sS32 _srcPitch
                                    current_transfer->width * 2,    // sU32 _numBytesX
                                    current_transfer->maxheight     // sU32 _numY
                                );

                                current_transfer->dst = (void *) (current_transfer->dstpitch + (uintptr_t) current_transfer->dst);
                                {
                                    uint16_t * restrict srcscreen;
                                    srcscreen = srcscreen1;
                                    srcscreen1 = srcscreen2;
                                    srcscreen2 = srcscreen;
                                }
                                {
                                    uint16_t * restrict dstscreen;
                                    dstscreen = dstscreen1;
                                    dstscreen1 = dstscreen2;
                                    dstscreen2 = dstscreen;
                                }
                            }

                            // last line:
                            // finish current A (Q2)
                            syscalls.qdma_wait(QDMA_2);

                            //   current C
                            blit_alpha_32_16_line_color(color, srcscreen1, dstscreen1, current_transfer->height * current_transfer->width);
                            //   finish previous D (Q0)
                            syscalls.qdma_wait(QDMA_0);
                            //   current D (Q0) (must finish on first line of next surface)
                            syscalls.qdma_copy2d(
                                QDMA_0,                         // sUI _qdmaNr
                                current_transfer->dst,          // void *_dst
                                current_transfer->dstpitch,     // sS32 _dstPitch
                                dstscreen1,                     // const void *_src
                                current_transfer->width * 2,    // sS32 _srcPitch
                                current_transfer->width * 2,    // sU32 _numBytesX
                                current_transfer->height        // sU32 _numY
                            );

                            {
                                uint16_t * restrict srcscreen;
                                srcscreen = srcscreen1;
                                srcscreen1 = srcscreen2;
                                srcscreen2 = srcscreen;
                            }
                            {
                                uint16_t * restrict dstscreen;
                                dstscreen = dstscreen1;
                                dstscreen1 = dstscreen2;
                                dstscreen2 = dstscreen;
                            }

                        }
                        else
                        {
                            uint32_t color2 = ((color & 0xf80000) >> 8) | ((color & 0xfc00) >> 5) | ((color & 0xf8) >> 3);
                            idma1_fill((uint32_t *) dstscreen1, color2 | (color2 << 16), 2 * cmd_queue[read_index].w + 3);

                            // no alpha
                            if (transfering_result)
                            {
                                syscalls.qdma_wait(QDMA_0);
                            }
                            transfering_result = 1;

                            // wait for idma
                            idma1_wait_all();

                            syscalls.qdma_copy2d(
                                QDMA_0,         // sUI _qdmaNr
                                (void *) ( (data->pitch * cmd_queue[read_index].dsty) + (2 * cmd_queue[read_index].dstx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                                // void *_dst
                                data->pitch,    // sS32 _dstPitch
                                dstscreen1,     // const void *_src
                                0,              // sS32 _srcPitch
                                2 * cmd_queue[read_index].w,
                                                // sU32 _numBytesX
                                cmd_queue[read_index].h
                                                // sU32 _numY
                            );

                            {
                                uint16_t * restrict dstscreen;
                                dstscreen = dstscreen1;
                                dstscreen1 = dstscreen2;
                                dstscreen2 = dstscreen;
                            }
                        }
                    }
                    break;
                case PARAM_RECT:
                    {
                        if (transfering_result)
                        {
                            syscalls.qdma_wait(QDMA_0);
                        }
                        transfering_result = 1;

                        int remaining = 0;
                        int nextchannel = 0xffff;
                        int curchannel;

                        int drawlines = cmd_queue[read_index].srcy;

                        if (drawlines & 1) remaining++;
                        if (drawlines & 2) remaining++;
                        if (drawlines & 4) remaining++;
                        if (drawlines & 8) remaining++;

                        uint32_t color = (cmd_queue[read_index].src_offset << 16) | cmd_queue[read_index].srcx;

                        *dstscreen2 = ((color & 0xf80000) >> 8) | ((color & 0xfc00) >> 5) | ((color & 0xf8) >> 3);

                        if (drawlines & 1)
                        {
                            // left line - vertical
                            curchannel = (nextchannel == 0xffff)?127:nextchannel - 1;
                            remaining--;
                            if (remaining == 0) curchannel = QDMA_0;
                            syscalls.qdma_link2d(
                                QDMA_0,         // sUI _qdmaNr
                                curchannel,     // sUI _paramNr
                                nextchannel,    // sUI _linkToParamNr
                                (void *) ( (data->pitch * cmd_queue[read_index].dsty) + (2 * cmd_queue[read_index].dstx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                                // void *_dst
                                data->pitch,    // sS32 _dstPitch
                                dstscreen2,     // const void *_src
                                0,              // sS32 _srcPitch
                                2,              // sU32 _numBytesX
                                cmd_queue[read_index].h
                                                // sU32 _numY
                            );
                            nextchannel = curchannel;
                        }

                        if (drawlines & 2)
                        {
                            // upper line - horizontal
                            curchannel = (nextchannel == 0xffff)?127:nextchannel - 1;
                            remaining--;
                            if (remaining == 0) curchannel = QDMA_0;
                            syscalls.qdma_link2d(
                                QDMA_0,         // sUI _qdmaNr
                                curchannel,     // sUI _paramNr
                                nextchannel,    // sUI _linkToParamNr
                                (void *) ( (data->pitch * cmd_queue[read_index].dsty) + (2 * cmd_queue[read_index].dstx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                                // void *_dst
                                2,              // sS32 _dstPitch
                                dstscreen2,     // const void *_src
                                0,              // sS32 _srcPitch
                                2,              // sU32 _numBytesX
                                cmd_queue[read_index].w
                                                // sU32 _numY
                            );
                            nextchannel = curchannel;
                        }

                        if (drawlines & 4)
                        {
                            // right line - vertical
                            curchannel = (nextchannel == 0xffff)?127:nextchannel - 1;
                            remaining--;
                            if (remaining == 0) curchannel = QDMA_0;
                            syscalls.qdma_link2d(
                                QDMA_0,         // sUI _qdmaNr
                                curchannel,     // sUI _paramNr
                                nextchannel,    // sUI _linkToParamNr
                                (void *) ( (data->pitch * cmd_queue[read_index].dsty) + (2 * (cmd_queue[read_index].dstx + cmd_queue[read_index].w - 1)) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                                // void *_dst
                                data->pitch,    // sS32 _dstPitch
                                dstscreen2,     // const void *_src
                                0,              // sS32 _srcPitch
                                2,              // sU32 _numBytesX
                                cmd_queue[read_index].h
                                                // sU32 _numY
                            );
                            nextchannel = curchannel;
                        }

                        if (drawlines & 8)
                        {
                            // bottom line - horizontal
                            curchannel = (nextchannel == 0xffff)?127:nextchannel - 1;
                            remaining--;
                            if (remaining == 0) curchannel = QDMA_0;
                            syscalls.qdma_link2d(
                                QDMA_0,         // sUI _qdmaNr
                                curchannel,     // sUI _paramNr
                                nextchannel,    // sUI _linkToParamNr
                                (void *) ( (data->pitch * (cmd_queue[read_index].dsty + cmd_queue[read_index].h - 1)) + (2 * cmd_queue[read_index].dstx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                                // void *_dst
                                2,              // sS32 _dstPitch
                                dstscreen2,     // const void *_src
                                0,              // sS32 _srcPitch
                                2,              // sU32 _numBytesX
                                cmd_queue[read_index].w
                                                // sU32 _numY
                            );
                            nextchannel = curchannel;
                        }
                    }
                    break;
                case PARAM_LINE:
                    if (transfering_result)
                    {
                        transfering_result = 0;
                        syscalls.qdma_wait(QDMA_0);
                    }

                    if (cmd_queue[read_index].srcy == cmd_queue[read_index].dsty)
                    {
                        // horizontal line
                        uint32_t color = (cmd_queue[read_index].w << 16) | cmd_queue[read_index].h;

                        *dstscreen2 = ((color & 0xf80000) >> 8) | ((color & 0xfc00) >> 5) | ((color & 0xf8) >> 3);

                        syscalls.qdma_copy2d(
                            QDMA_0,         // sUI _qdmaNr
                            (void *) ( (data->pitch * cmd_queue[read_index].srcy) + (2 * cmd_queue[read_index].srcx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                            // void *_dst
                            2,              // sS32 _dstPitch
                            dstscreen2,     // const void *_src
                            0,              // sS32 _srcPitch
                            2,              // sU32 _numBytesX
                            cmd_queue[read_index].dstx - cmd_queue[read_index].srcx
                                            // sU32 _numY
                        );

                        transfering_result = 1;
                    }
                    else if (cmd_queue[read_index].srcx == cmd_queue[read_index].dstx)
                    {
                        // vertical line

                        uint32_t color = (cmd_queue[read_index].w << 16) | cmd_queue[read_index].h;

                        *dstscreen2 = ((color & 0xf80000) >> 8) | ((color & 0xfc00) >> 5) | ((color & 0xf8) >> 3);

                        syscalls.qdma_copy2d(
                            QDMA_0,         // sUI _qdmaNr
                            (void *) ( (data->pitch * cmd_queue[read_index].srcy) + (2 * cmd_queue[read_index].srcx) + (uintptr_t) data->pixels[cmd_queue[read_index].dst_screen] ),
                                            // void *_dst
                            data->pitch,    // sS32 _dstPitch
                            dstscreen2,     // const void *_src
                            0,              // sS32 _srcPitch
                            2,              // sU32 _numBytesX
                            cmd_queue[read_index].dsty - cmd_queue[read_index].srcy
                                            // sU32 _numY
                        );

                        transfering_result = 1;
                    }
                    else
                    {
                        draw_line_16(data, &(cmd_queue[read_index]));
                    }
                    break;
            }

            read_index = (read_index + 1) & (CMD_QUEUE_LENGTH - 1);
            data->cmd_queue_read = read_index;
        } while (read_index != data->cmd_queue_write);
    };
}

#undef MAX_PIXELS_PER_LINE

/*
 * 1 pixel:
 * 2 * (srcalpha + srcscreen + dstscreen)
 * 2 * (4 + 4 + 4)
 * 24 bytes per pixes
 * (24*1024-64-32-32) / 24 = max 1018 pixels at once
 */
#define MAX_PIXELS_PER_LINE 1018
// should be even number

static void tiles_run_32(dsp_screen *data)
{
    // todo:
    int read_index = data->cmd_queue_read;
    while (1)
    {
        while (read_index == data->cmd_queue_write)
        {
            if (data->quit) return;
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
            asm(" NOP\n");
        };

        do {
            read_index = (read_index + 1) & (CMD_QUEUE_LENGTH - 1);
            data->cmd_queue_read = read_index;
        } while (read_index != data->cmd_queue_write);
    };
}

#undef MAX_PIXELS_PER_LINE

void tiles_run(dsp_screen *data)
{
    if (data->bytes_per_pixel == 4)
    {
        tiles_run_32(data);
    }
    else
    {
        tiles_run_16(data);
    }

/*
 * A) copy dst to localdst
 * B) copy src to localsrc
 * C) set localdst2 using localsrc and localdst (depends on A,B)
 * D) copy localdst2 to dst
 *
 *
 *
 * NONALPHA:
 * if (PREV D)
 *   finish PREV D (Q0)
 * if (! NEXT A,B)
 *   current D (??? link NEXT NONALPHA ???) (Q0)
 * else if (NEXT A depends on CURR D)
 *   NEXT B (Q1) | current D, NEXT A (Q2) (linked !!!)
 * else
 *   current D (Q0) | NEXT B,A (Q2) (linked !!!)
 *
 * ALPHA:
 * first line:
 * if (CURR A,B)
 *   // nothing
 * else if (PREV D)
 *   first B (Q1)
 *   finish PREV D (Q0)
 *   first A (Q2)
 * else
 *   first B,A (linked or parallel) (Q2 or Q1,Q2)
 *
 * line loop:
 * finish current B,A (Q1,Q2)
 * next B,A (linked or parallel) (Q2 or Q1,Q2) | current C
 * finish previous D (Q0)
 * current D (Q0)
 *
 * last line:
 * finish current B,A (Q1,Q2)
 * if (! NEXT A,B)
 *   current C
 *   finish previous D (Q0)
 *   current D (Q0) (must finish on first line of next surface) (??? link NEXT NONALPHA ???)
 * else if (NEXT A depends on CURR D)
 *   NEXT B (Q1) | current C
 *   finish previous D (Q0)
 *   current D, NEXT A (Q2) (linked !!!) (must finish on first line of next surface)
 * else
 *   NEXT B,A (linked or parallel) (Q2 or Q1,Q2) | current C
 *   finish previous D (Q0)
 *   current D (Q0) (must finish on first line of next surface)
 *
 * */
}

