/*
 * Copyright (C) 2016  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "SDL.h"
#include <stdlib.h>
#include <stdint.h>
#include "dsptiles.h"
// dsp - c64
#include "../include/inc_libc64_mini.h"
#include "tiles_dsp.h"


static int dsp_initialized = 0;
static int screen_w, screen_h, screen_bytes_per_pixel, screen_pitch, screen_active;
static void *screen_pixels[2];
static int flip_started, flip_readindex, flip_writeindex;
static void *flip_srcpixels;

static dsp_screen *dsp_data;
static cmd_params *cmd_queue;

static dsp_component_id_t compid_tiles_dsp;
static dsp_mem_region_t scratchmem, dspdatamem, surfacemem, surfacedatamem, screenmem;
static dsp_mspace_handle_t surfaceheap, surfacedataheap;


static int is_screen_format(SDL_Surface* screen)
{
    if (screen->format->BitsPerPixel == 16)
    {
        if ((screen->format->Rmask == 0xf800) &&
            (screen->format->Gmask == 0x07e0) &&
            (screen->format->Bmask == 0x001f) &&
            (screen->format->Amask == 0)
        ) return 1;
    }
    else if (screen->format->BitsPerPixel == 32)
    {
        if ((screen->format->Rmask == 0x00ff0000) &&
            (screen->format->Gmask == 0x0000ff00) &&
            (screen->format->Bmask == 0x000000ff) &&
            (screen->format->Amask == 0)
        ) return 1;
    }

    return 0;
}

void dsptiles_quit(void)
{
    if (!dsp_initialized) return;

    // wait for dsp to finish work
    int write_index = dsp_data->cmd_queue_write;
    while (write_index != dsp_data->cmd_queue_read) SDL_Delay(0);

    dsp_data->quit = 1;

    // finish dsp procedure
    dsp_rpc_recv(NULL);

    dsp_initialized = 0;

    dsp_mspace_destroy(surfacedataheap);
    dsp_mspace_destroy(surfaceheap);

    dsp_shm_free(screenmem);
    dsp_shm_free(surfacedatamem);
    dsp_l1sram_free(surfacemem);
    dsp_l1sram_free(dspdatamem);

    dsp_close();
}

int dsptiles_init(SDL_Surface* screen)
{
    if (dsp_initialized) return -1;

    screen_pixels[0] = NULL;
    screen_pixels[1] = NULL;
    flip_started = 0;

    if (screen == NULL) return -1;

    if (!is_screen_format(screen)) return -1;

    screen_w = screen->w;
    screen_h = screen->h;
    screen_bytes_per_pixel = screen->format->BytesPerPixel;

    int w4 = (screen_w + 3) & ~3;

    screen_pitch = w4 * screen_bytes_per_pixel;
    screen_active = 0;


    system("sudo /usr/pandora/scripts/op_dsp_c64.sh");

    int ret = dsp_open();
    if (0 != ret) return -1;

    if (0 != dsp_component_load(NULL, COMPONENT_NAME_TILES_DSP, &compid_tiles_dsp))
    {
        dsp_close();
        return -1;
    }

    scratchmem = dsp_l1sram_scratchbuffer_get();
    if (0 == scratchmem.size)
    {
        dsp_close();
        return -1;
    }

    dspdatamem = dsp_l1sram_alloc(63 + sizeof(dsp_screen));
    if (0 == dspdatamem.size)
    {
        dsp_close();
        return -1;
    }

    surfacemem = dsp_l1sram_alloc(7 * 1024);
    if (0 == surfacemem.size)
    {
        dsp_l1sram_free(dspdatamem);
        dsp_close();
        return -1;
    }

    surfacedatamem = dsp_shm_alloc(DSP_CACHE_W, 32*1024*1024);
    if (0 == surfacedatamem.size)
    {
        dsp_l1sram_free(surfacemem);
        dsp_l1sram_free(dspdatamem);
        dsp_close();
        return -1;
    }

    screenmem = dsp_shm_alloc(DSP_CACHE_R, 2 * screen_pitch * screen_h);
    if (0 == screenmem.size)
    {
        dsp_shm_free(surfacedatamem);
        dsp_l1sram_free(surfacemem);
        dsp_l1sram_free(dspdatamem);
        dsp_close();
        return -1;
    }

    screen_pixels[0] = (void *) screenmem.virt_addr;
    screen_pixels[1] = (void *) (screen_pitch * screen_h + (uintptr_t)screen_pixels[0]);

    dsp_data = (dsp_screen *) ((dspdatamem.virt_addr + 63) & ~63);

    cmd_queue = (cmd_params *) scratchmem.virt_addr;
    dsp_data->cmd_queue_write = 0;
    dsp_data->w = screen_w;
    dsp_data->h = screen_h;
    dsp_data->bytes_per_pixel = screen_bytes_per_pixel;
    dsp_data->pitch = screen_pitch;
    dsp_data->pixels[0] = (void *) dsp_virt_to_phys(screen_pixels[0]);
    dsp_data->pixels[1] = (void *) dsp_virt_to_phys(screen_pixels[1]);
    dsp_data->quit = 0;
    dsp_data->surfaces_base = (uint8_t *) dsp_physgpp_to_physdsp(surfacemem.phys_addr);
    dsp_data->surfacedata = (void *) dsp_physgpp_to_physdsp(surfacedatamem.phys_addr);
    dsp_data->surfacedata_size = surfacedatamem.size;
    dsp_data->cmd_queue_read = 0;

    // run initialize on dsp (invalidate caches, initialize dma, ...)
    {
        dsp_msg_t msg;

        DSP_MSG_INIT(&msg, compid_tiles_dsp, TILES_DSP_INIT_TILES,
                     dsp_virt_to_phys(dsp_data),           // argument #1
                     0            // argument #2
                    );

        ret = dsp_rpc(&msg, NULL);

        if (0 != ret)
        {
            dsp_shm_free(screenmem);
            dsp_shm_free(surfacedatamem);
            dsp_l1sram_free(surfacemem);
            dsp_l1sram_free(dspdatamem);
            dsp_close();
            return -1;
        }
    }

    // run procedure on dsp
    {
        dsp_msg_t msg;

        DSP_MSG_INIT(&msg, compid_tiles_dsp, TILES_DSP_RUN_TILES,
                     dsp_virt_to_phys(dsp_data),           // argument #1
                     0            // argument #2
                    );

        ret = dsp_rpc_send(&msg);

        if (0 != ret)
        {
            dsp_shm_free(screenmem);
            dsp_shm_free(surfacedatamem);
            dsp_l1sram_free(surfacemem);
            dsp_l1sram_free(dspdatamem);
            dsp_close();
            return -1;
        }
    }

    surfaceheap = dsp_mspace_create (surfacemem.virt_addr, 7 * 1024);
    surfacedataheap = dsp_mspace_create (surfacedatamem.virt_addr, 32 * 1024 * 1024);

    dsp_initialized = 1;
    atexit(dsptiles_quit);

    return 0;
}


dsptiles_surface *dsptiles_add_surface(SDL_Surface* surface)
{
    if (!dsp_initialized) return NULL;

    if (surface->format->Amask == 0)
    {
        if (screen_bytes_per_pixel != surface->format->BytesPerPixel) return NULL;

        if (!is_screen_format(surface)) return NULL;
    }
    else
    {
        if (surface->format->BitsPerPixel == 32)
        {
            if ((surface->format->Rmask != 0x00ff0000) ||
                (surface->format->Gmask != 0x0000ff00) ||
                (surface->format->Bmask != 0x000000ff) ||
                (surface->format->Amask != 0xff000000)
            ) return NULL;
        }
        else return NULL;
    }

    int w4 = (surface->w + 3) & ~3;
    int pitch = w4 * surface->format->BytesPerPixel;

    void *pixels = dsp_mspace_malloc(surfacedataheap, pitch * surface->h);
    if (pixels == NULL) return NULL;

    dsp_surface *dsp_ret = (dsp_surface *) dsp_mspace_malloc(surfaceheap, sizeof(dsp_surface));
    if (dsp_ret == NULL)
    {
        dsp_mspace_free(surfacedataheap, pixels);
        return NULL;
    }

    dsptiles_surface *ret = (dsptiles_surface *)malloc(sizeof(dsptiles_surface));
    if (ret == NULL)
    {
        dsp_mspace_free(surfaceheap, dsp_ret);
        dsp_mspace_free(surfacedataheap, pixels);
        return NULL;
    }

    ret->w = surface->w;
    ret->h = surface->h;
    ret->bytes_per_pixel = surface->format->BytesPerPixel;
    ret->has_alpha = (surface->format->Amask)?1:0;
    ret->pitch = pitch;
    ret->pixels = pixels;
    ret->dsp = dsp_ret;
    ret->dsp_offset = ((uintptr_t)dsp_ret) - surfacemem.virt_addr;

    dsp_ret->w = surface->w;
    dsp_ret->h = surface->h;
    dsp_ret->bytes_per_pixel = surface->format->BytesPerPixel;
    dsp_ret->has_alpha = (surface->format->Amask)?1:0;
    dsp_ret->pitch = pitch;
    dsp_ret->pixels = (void *) dsp_virt_to_phys(pixels);


    void *srcpixels = surface->pixels;
    void *dstpixels = pixels;

    int copywidth = surface->w * surface->format->BytesPerPixel;
    int y;
    for (y = surface->h; y != 0; y--)
    {
        memcpy(dstpixels, srcpixels, copywidth);
        srcpixels = (void *) (surface->pitch + (uintptr_t)srcpixels);
        dstpixels = (void *) (pitch + (uintptr_t)dstpixels);
    }


    return ret;
}


int dsptiles_free_surface(dsptiles_surface *surface)
{
    if (!dsp_initialized) return -1;

    if (surface == NULL) return 0;

    // wait for dsp to finish work
    int write_index = dsp_data->cmd_queue_write;
    while (write_index != dsp_data->cmd_queue_read) SDL_Delay(0);

    dsp_mspace_free(surfaceheap, surface->dsp);
    dsp_mspace_free(surfacedataheap, surface->pixels);
    free(surface);

    return 0;
}


int dsptiles_blit_surface(dsptiles_surface *src, SDL_Rect *srcrect, SDL_Rect *dstrect)
{
    if (!dsp_initialized) return -1;

    // a bit of a hack
    if (flip_started) return 0;

    SDL_Rect srcpos, dstpos;

    if (dstrect != NULL)
    {
        dstpos.x = dstrect->x;
        dstpos.y = dstrect->y;
    }
    else
    {
        dstpos.x = dstpos.y = 0;
    }

    if (srcrect != NULL)
    {
        srcpos = *srcrect;

        // clip src rectangle against src
        if (srcpos.x < 0)
        {
            dstpos.x -= srcpos.x;
            srcpos.w += srcpos.x;
            srcpos.x = 0;
            if (((Sint16)srcpos.w) <= 0) return 0;
        }

        if (srcpos.w > (src->w - srcpos.x))
        {
            srcpos.w = src->w - srcpos.x;
            if (((Sint16)srcpos.w) <= 0) return 0;
        }

        if (srcpos.y < 0)
        {
            dstpos.y -= srcpos.y;
            srcpos.h += srcpos.y;
            srcpos.y = 0;
            if (((Sint16)srcpos.h) <= 0) return 0;
        }

        if (srcpos.h > (src->h - srcpos.y))
        {
            srcpos.h = src->h - srcpos.y;
            if (((Sint16)srcpos.h) <= 0) return 0;
        }
    }
    else
    {
        srcpos.x = srcpos.y = 0;
        srcpos.w = src->w;
        srcpos.h = src->h;
    }

    // clip dst rectangle against dst
    {
        if (dstpos.x < 0)
        {
            srcpos.x -= dstpos.x;
            srcpos.w += dstpos.x;
            dstpos.x = 0;
            if (((Sint16)srcpos.w) <= 0) return 0;
        }

        if (srcpos.w > (screen_w - dstpos.x))
        {
            srcpos.w = screen_w - dstpos.x;
            if (((Sint16)srcpos.w) <= 0) return 0;
        }

        if (dstpos.y < 0)
        {
            srcpos.y -= dstpos.y;
            srcpos.h += dstpos.y;
            dstpos.y = 0;
            if (((Sint16)srcpos.h) <= 0) return 0;
        }

        if (srcpos.h > (screen_h - dstpos.y))
        {
            srcpos.h = screen_h - dstpos.y;
            if (((Sint16)srcpos.h) <= 0) return 0;
        }
    }

    // insert blit into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full

    cmd_queue[write_index].type = PARAM_BLIT;
    cmd_queue[write_index].dst_screen = screen_active;
    cmd_queue[write_index].src_offset = src->dsp_offset;
    cmd_queue[write_index].srcx = srcpos.x;
    cmd_queue[write_index].srcy = srcpos.y;
    cmd_queue[write_index].w = srcpos.w;
    cmd_queue[write_index].h = srcpos.h;
    cmd_queue[write_index].dstx = dstpos.x;
    cmd_queue[write_index].dsty = dstpos.y;

    dsp_data->cmd_queue_write = write_index;

    return 0;
}

int dsptiles_rectangle_filled(SDL_Rect *dstrect, uint32_t color, int alpha)
{
    if (!dsp_initialized) return -1;

    // a bit of a hack
    if (flip_started) return 0;

    if (alpha == 0) return 0;

    SDL_Rect dstpos;

    if (dstrect != NULL)
    {
        dstpos = *dstrect;

        // clip rectangle
        if (dstpos.x >= screen_w) return 0;
        if (dstpos.x + dstpos.w <= 0) return 0;
        if (dstpos.x + dstpos.w > screen_w) dstpos.w = screen_w - dstpos.x;

        if (dstpos.x < 0) {
            dstpos.w += dstpos.x;
            dstpos.x = 0;
        }

        if (dstpos.y >= screen_h) return 0;
        if (dstpos.y + dstpos.h <= 0) return 0;
        if (dstpos.y + dstpos.h > screen_h) dstpos.h = screen_h - dstpos.y;

        if (dstpos.y < 0) {
            dstpos.h += dstpos.y;
            dstpos.y = 0;
        }
    }
    else
    {
        dstpos.x = dstpos.y = 0;
        dstpos.w = screen_w;
        dstpos.h = screen_h;
    }

    color = (color & 0x00ffffff) | (alpha << 24);

    // insert filled rectangle into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full

    cmd_queue[write_index].type = PARAM_FRECT;
    cmd_queue[write_index].dst_screen = screen_active;
    cmd_queue[write_index].src_offset = (color >> 16);
    cmd_queue[write_index].srcx = (color & 0xffff);
    cmd_queue[write_index].srcy = 0;
    cmd_queue[write_index].w = dstpos.w;
    cmd_queue[write_index].h = dstpos.h;
    cmd_queue[write_index].dstx = dstpos.x;
    cmd_queue[write_index].dsty = dstpos.y;

    dsp_data->cmd_queue_write = write_index;

    return 0;
}

int dsptiles_rectangle(SDL_Rect *dstrect, uint32_t color)
{
    if (!dsp_initialized) return -1;

    // a bit of a hack
    if (flip_started) return 0;

    SDL_Rect dstpos;
    int drawlines = 0x0f;

    if (dstrect != NULL)
    {
        dstpos = *dstrect;

        // clip rectangle
        if (dstpos.x >= screen_w) return 0;
        if (dstpos.x + dstpos.w <= 0) return 0;
        if (dstpos.x + dstpos.w > screen_w) {
            dstpos.w = screen_w - dstpos.x;
            drawlines &= ~4;
        }

        if (dstpos.x < 0) {
            dstpos.w += dstpos.x;
            dstpos.x = 0;
            drawlines &= ~1;
        }

        if (dstpos.y >= screen_h) return 0;
        if (dstpos.y + dstpos.h <= 0) return 0;
        if (dstpos.y + dstpos.h > screen_h) {
            dstpos.h = screen_h - dstpos.y;
            drawlines &= ~8;
        }

        if (dstpos.y < 0) {
            dstpos.h += dstpos.y;
            dstpos.y = 0;
            drawlines &= ~2;
        }

        if (drawlines == 0) return 0;
    }
    else
    {
        dstpos.x = dstpos.y = 0;
        dstpos.w = screen_w;
        dstpos.h = screen_h;
    }

    color = (color & 0x00ffffff);

    // insert rectangle into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full

    cmd_queue[write_index].type = PARAM_RECT;
    cmd_queue[write_index].dst_screen = screen_active;
    cmd_queue[write_index].src_offset = (color >> 16);
    cmd_queue[write_index].srcx = (color & 0xffff);
    cmd_queue[write_index].srcy = drawlines;
    cmd_queue[write_index].w = dstpos.w;
    cmd_queue[write_index].h = dstpos.h;
    cmd_queue[write_index].dstx = dstpos.x;
    cmd_queue[write_index].dsty = dstpos.y;

    dsp_data->cmd_queue_write = write_index;

    return 0;
}

int dsptiles_line(int x1, int y1, int x2, int y2, uint32_t color)
{
    if (!dsp_initialized) return -1;

    // a bit of a hack
    if (flip_started) return 0;

    int ty, tx;
    if (y2 < y1)
    {
        ty = y1;
        y1 = y2;
        y2 = ty;

        tx = x1;
        x1 = x2;
        x2 = tx;
    }

    // clip line
    if (y2 < 0) return 0;
    if (y1 >= screen_h) return 0;

    if (y1 == y2)
    {
        if (x2 < x1)
        {
            tx = x1;
            x1 = x2;
            x2 = tx;
        }

        if (x2 < 0) return 0;
        if (x1 >= screen_w) return 0;
        if (x2 >= screen_w) x2 = screen_w - 1;
        if (x1 < 0) x1 = 0;
    }
    else if (x1 == x2)
    {
        if (x1 < 0) return 0;
        if (x1 >= screen_w) return 0;

        if (y1 < 0) y1 = 0;
        if (y2 >= screen_h) y2 = screen_h - 1;
    }
    else
    {
        if (y1 < 0)
        {
            x1 = x1 + (0 - y1) * ( ((float)(x2 - x1)) / ((float)(y2 - y1)) );
            y1 = 0;
        }
        if (y2 >= screen_h)
        {
            x2 = x1 + (screen_h - 1 - y1) * ( ((float)(x2 - x1)) / ((float)(y2 - y1)) );
            y2 = screen_h - 1;
        }

        if (x1 < x2)
        {
            if (x2 < 0) return 0;
            if (x1 >= screen_w) return 0;

            if (x1 < 0)
            {
                y1 = y1 + (0 - x1) * ( ((float)(y2 - y1)) / ((float)(x2 - x1)) );
                x1 = 0;
            }
            if (x2 >= screen_w)
            {
                y2 = y1 + (screen_w - 1 - x1) * ( ((float)(y2 - y1)) / ((float)(x2 - x1)) );
                x2 = screen_w - 1;
            }
        }
        else
        {
            if (x1 < 0) return 0;
            if (x2 >= screen_w) return 0;

            if (x2 < 0)
            {
                y2 = y1 + (0 - x1) * ( ((float)(y2 - y1)) / ((float)(x2 - x1)) );
                x2 = 0;
            }
            if (x1 >= screen_w)
            {
                y1 = y1 + (screen_w - 1 - x1) * ( ((float)(y2 - y1)) / ((float)(x2 - x1)) );
                x1 = screen_w - 1;
            }
        }
    }

    color = (color & 0x00ffffff);

    // insert line into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full

    cmd_queue[write_index].type = PARAM_LINE;
    cmd_queue[write_index].dst_screen = screen_active;
    cmd_queue[write_index].src_offset = 0;
    cmd_queue[write_index].srcx = x1;
    cmd_queue[write_index].srcy = y1;
    cmd_queue[write_index].w = (color >> 16);
    cmd_queue[write_index].h = (color & 0xffff);
    cmd_queue[write_index].dstx = x2;
    cmd_queue[write_index].dsty = y2;

    dsp_data->cmd_queue_write = write_index;

    return 0;
}

int dsptiles_flip(SDL_Surface* screen)
{
    if (!dsp_initialized) return -1;

    if (screen == NULL) return -1;
    if ((screen_w != screen->w) || (screen_h != screen->h)) return -1;
    if (screen_bytes_per_pixel != screen->format->BytesPerPixel) return -1;
    if (!is_screen_format(screen)) return -1;


    // insert flush into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full

    cmd_queue[write_index].type = PARAM_FLUSH;
    cmd_queue[write_index].dst_screen = screen_active;

    dsp_data->cmd_queue_write = write_index;

    // invalidate screen memory
    dsp_cache_inv_virt(screen_pixels[screen_active], screen_pitch * screen_h);

    flip_started = 0;

    SDL_LockSurface(screen);

    // wait for dsp to finish work
    while (write_index != dsp_data->cmd_queue_read) SDL_Delay(0);



    void *srcpixels = screen_pixels[screen_active];
    screen_active = 1 - screen_active;
    void *dstpixels = screen->pixels;

    int copywidth = screen_w * screen_bytes_per_pixel;
    int y;
    for (y = screen_h; y != 0; y--)
    {
        memcpy(dstpixels, srcpixels, copywidth);
        srcpixels = (void *) (screen_pitch + (uintptr_t)srcpixels);
        dstpixels = (void *) (screen->pitch + (uintptr_t)dstpixels);
    }

    SDL_UnlockSurface(screen);

    return SDL_Flip(screen);
}

int dsptiles_flip_start(SDL_Surface* screen)
{
    if (!dsp_initialized) return -1;

    if (screen == NULL) return -1;
    if ((screen_w != screen->w) || (screen_h != screen->h)) return -1;
    if (screen_bytes_per_pixel != screen->format->BytesPerPixel) return -1;
    if (!is_screen_format(screen)) return -1;


    // insert flush into command queue
    int write_index = (dsp_data->cmd_queue_write + 1) & (CMD_QUEUE_LENGTH - 1);
    while (write_index == dsp_data->cmd_queue_read) SDL_Delay(0); // wait while queue is full
    flip_readindex = dsp_data->cmd_queue_read;
    flip_writeindex = write_index;

    cmd_queue[write_index].type = PARAM_FLUSH;
    cmd_queue[write_index].dst_screen = screen_active;

    dsp_data->cmd_queue_write = write_index;

    // invalidate screen memory
    dsp_cache_inv_virt(screen_pixels[screen_active], screen_pitch * screen_h);

    flip_srcpixels = screen_pixels[screen_active];
    screen_active = 1 - screen_active;

    flip_started = 1;

    return 0;
}

int dsptiles_flip_finish(SDL_Surface* screen)
{
    if (!dsp_initialized) return -1;

    if (screen == NULL) return -1;
    if ((screen_w != screen->w) || (screen_h != screen->h)) return -1;
    if (screen_bytes_per_pixel != screen->format->BytesPerPixel) return -1;
    if (!is_screen_format(screen)) return -1;

    if (!flip_started) return 0;
    flip_started = 0;

    SDL_LockSurface(screen);

    // wait for dsp to finish work
    int write_index = dsp_data->cmd_queue_write;
    while (write_index != dsp_data->cmd_queue_read) SDL_Delay(0);



    void *srcpixels = flip_srcpixels;
    void *dstpixels = screen->pixels;

    int copywidth = screen_w * screen_bytes_per_pixel;
    int y;
    for (y = screen_h; y != 0; y--)
    {
        memcpy(dstpixels, srcpixels, copywidth);
        srcpixels = (void *) (screen_pitch + (uintptr_t)srcpixels);
        dstpixels = (void *) (screen->pitch + (uintptr_t)dstpixels);
    }

    SDL_UnlockSurface(screen);

    return SDL_Flip(screen);
}

